# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements support for notifications/alerts sent by G90 alarm panel.
"""
import logging
from typing import (
    Optional, Tuple, Callable
)
import asyncio
from asyncio.transports import BaseTransport
from asyncio.protocols import DatagramProtocol

from ..notifications.base import G90NotificationsBase
from ..notifications.protocol import G90NotificationProtocol

_LOGGER = logging.getLogger(__name__)


class G90LocalNotifications(G90NotificationsBase, DatagramProtocol):
    """
    Implements support for notifications/alerts sent by alarm panel.

    There is a basic check to ensure only notifications/alerts from the correct
    device are processed - the check uses the host and port of the device, and
    the device ID (GUID) that is set by the ancestor class that implements the
    commands (e.g. :class:`G90Alarm`). The latter to work correctly needs a
    command to be performed first, one that fetches device GUID and then stores
    it using :attr:`.device_id` (e.g. :meth:`G90Alarm.get_host_info`).

    :param protocol_factory: A callable that returns a new instance of the
     :class:`G90NotificationProtocol` class.
    :param port: The port on which the device is listening for notifications.
    :param host: The host on which the device is listening for notifications.
    :param local_port: The port on which the local host is listening for
     notifications.
    :param local_host: The host on which the local host is listening for
     notifications.
    """
    def __init__(  # pylint:disable=too-many-arguments
        self, protocol_factory: Callable[[], G90NotificationProtocol],
        port: int, host: str, local_port: int, local_host: str,
    ):
        super().__init__(protocol_factory)

        self._host = host
        self._port = port
        self._notifications_local_host = local_host
        self._notifications_local_port = local_port

    # Implementation of datagram protocol,
    # https://docs.python.org/3/library/asyncio-protocol.html#datagram-protocols
    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection from the device is made.
        """

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Same but when the connection is lost.
        """

    def datagram_received(
        self, data: bytes, addr: Tuple[str, int]
    ) -> None:
        """
        Invoked when datagram is received from the device.
        """
        if self._host and self._host != addr[0]:
            _LOGGER.error(
                "Received notification/alert from wrong host '%s',"
                " expected from '%s'",
                addr[0], self._host
            )
            return

        self.set_last_device_packet_time()

        _LOGGER.debug('Received device message from %s:%s: %s',
                      addr[0], addr[1], data)

        self.handle(data)

    async def listen(self) -> None:
        """
        Listens for notifications/alerts from the device.
        """
        loop = asyncio.get_running_loop()

        _LOGGER.debug('Creating UDP endpoint for %s:%s',
                      self._notifications_local_host,
                      self._notifications_local_port)
        (self._transport,
         _protocol) = await loop.create_datagram_endpoint(
            lambda: self,
            local_addr=(
                self._notifications_local_host, self._notifications_local_port
            ))



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Definies different constants for G90 alarm panel.
"""
from __future__ import annotations
from enum import IntEnum
from typing import Optional

REMOTE_PORT = 12368
REMOTE_TARGETED_DISCOVERY_PORT = 12900
LOCAL_TARGETED_DISCOVERY_PORT = 12901
LOCAL_NOTIFICATIONS_HOST = '0.0.0.0'
LOCAL_NOTIFICATIONS_PORT = 12901
CLOUD_NOTIFICATIONS_HOST = '0.0.0.0'
CLOUD_NOTIFICATIONS_PORT = 5678
REMOTE_CLOUD_HOST = '47.88.7.61'
REMOTE_CLOUD_PORT = 5678

CMD_PAGE_SIZE = 10


class G90Commands(IntEnum):
    """
    Defines the alarm panel commands and their codes.

    The list consists of the entities known so far, and does not pretend to be
    comprehensive or complete.
    """

    def __new__(cls, value: int, doc: Optional[str] = None) -> G90Commands:
        """
        Allows to set the docstring along with the value to enum entry.
        """
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.__doc__ = doc
        return obj

    NONE = (0, """
        Pseudo command, to be used for proper typing with subclasses of
        `G90BaseCommand` invoking its constructor but implementing special
        processing
    """)

    # Host status
    GETHOSTSTATUS = (100, 'Get host status')
    SETHOSTSTATUS = (101, 'Set host status')
    # Host info
    GETHOSTINFO = 206
    # History
    GETHISTORY = 200
    # Sensors
    GETSENSORLIST = (102, """
        Get list of sensors

        .. note:: Paginated command, see :py:class:`.G90PaginatedResult`
    """)
    SETSINGLESENSOR = 103
    DELSENSOR = 131
    ADDSENSOR = 156
    LEARNSENSOR = 157
    CANCELLEARNSENSOR = 163
    DELALLSENSORS = 202
    # Switches (relays)
    ADDDEVICE = 134
    REGDEVICE = 135
    DELDEVICE = 136
    CONTROLDEVICE = 137
    GETDEVICELIST = (138, """
        Get list of devices (switches)

        .. note:: Paginated command, see :py:class:`.G90PaginatedResult`
    """)
    GETSINGLEDEVICE = 139
    SETSINGLEDEVICE = 140
    DELALLDEVICES = 203
    # Host config
    GETHOSTCONFIG = 106
    SETHOSTCONFIG = 107
    SETALMPHONE = 108
    SETAUTOARM = 109
    # Wireless sirens
    GETSIREN = 110
    SETSIREN = 111
    # Alarm phones, notifications
    GETALMPHONE = 114
    GETAUTOARM = 115
    SETNOTICEFLAG = 116
    GETNOTICEFLAG = 117
    # Factory reset
    SETFACTORY = 118
    GETALARM = 119
    # Rooms
    SETROOMINFO = 141
    GETROOMINFO = 142
    ADDROOM = 158
    DELROOM = 159
    # Scenes
    ADDSCENE = 143
    DELSCENE = 144
    CTLSCENE = 145
    GETSCENELIST = (146, """
        Get list of scenes

        .. note:: Paginated command, see :py:class:`.G90PaginatedResult`
    """)
    GETSINGLESCENE = 147
    SETSINGLESCENE = 148
    GETROOMANDSCENE = 149
    DELALLSCENES = 204
    # IFTTT (scenarios)
    ADDIFTTT = 150
    DELIFTTT = 151
    GETIFTTTLIST = (152, """
        Get list of if-then-else scenarios

        .. note:: Paginated command, see :py:class:`.G90PaginatedResult`
    """)
    GETSINGLEIFTTT = 153
    SETSINGLEIFTTT = 154
    IFTTTREQTIMERID = 164
    DELALLIFTTT = 205
    # Data CRC
    GETUSERDATACRC = 160
    # Fingerprint scanners
    GETFPLOCKLIST = (165, """
        Get list of fingerprint scanners

        .. note:: Paginated command, see :py:class:`.G90PaginatedResult`
    """)
    SETFPLOCKNAME = 166
    GETFPLOCKUSERNAME = 167
    SETFPLOCKUSERNAME = 168
    DELALLLOCK = 223
    # Miscellaneous
    GETAPINFO = 212
    PINGBYGPRS = 218
    PING = 219


class G90MessageTypes(IntEnum):
    """
    Defines message types (codes) from messages coming from the alarm panel.
    """
    NOTIFICATION = 170
    ALERT = 208


class G90NotificationTypes(IntEnum):
    """
    Defines types of notifications sent by the alarm panel.
    """
    ARM_DISARM = 1
    SENSOR_ADDED = 4
    SENSOR_ACTIVITY = 5
    DOOR_OPEN_WHEN_ARMING = 6
    FIRMWARE_UPDATING = 8


class G90ArmDisarmTypes(IntEnum):
    """
    Defines arm/disarm states of the device, applicable both for setting device
    state and one the device sends in notification messages.
    """
    ARM_AWAY = 1
    ARM_HOME = 2
    DISARM = 3
    ALARMED = 4


class G90AlertTypes(IntEnum):
    """
    Defines types of alerts sent by the alarm panel.
    """
    HOST_SOS = 1
    STATE_CHANGE = 2
    ALARM = 3
    SENSOR_ACTIVITY = 4
    # Retained for compatibility, deprecated
    DOOR_OPEN_CLOSE = 4


class G90AlertSources(IntEnum):
    """
    Defines possible sources of the alert sent by the panel.
    """
    DEVICE = 0
    SENSOR = 1
    TAMPER = 3
    REMOTE = 10
    RFID = 11
    DOORBELL = 12
    FINGERPRINT = 15


class G90AlertStates(IntEnum):
    """
    Defines possible states of the alert sent by the panel.
    """
    DOOR_CLOSE = 0
    DOOR_OPEN = 1
    SOS = 2
    TAMPER = 3
    LOW_BATTERY = 4
    ALARM = 254


class G90AlertStateChangeTypes(IntEnum):
    """
    Defines types of alert for device state changes.
    """
    AC_POWER_FAILURE = 1
    AC_POWER_RECOVER = 2
    DISARM = 3
    ARM_AWAY = 4
    ARM_HOME = 5
    LOW_BATTERY = 6
    WIFI_CONNECTED = 7
    WIFI_DISCONNECTED = 8


class G90HistoryStates(IntEnum):
    """
    Defines possible states for history entities.
    """
    DOOR_CLOSE = 1
    DOOR_OPEN = 2
    TAMPER = 3
    ALARM = 4
    AC_POWER_FAILURE = 5
    AC_POWER_RECOVER = 6
    DISARM = 7
    ARM_AWAY = 8
    ARM_HOME = 9
    LOW_BATTERY = 10
    WIFI_CONNECTED = 11
    WIFI_DISCONNECTED = 12
    REMOTE_BUTTON_ARM_AWAY = 13
    REMOTE_BUTTON_ARM_HOME = 14
    REMOTE_BUTTON_DISARM = 15
    REMOTE_BUTTON_SOS = 16


class G90RemoteButtonStates(IntEnum):
    """
    Defines possible states for remote control buttons.
    """
    ARM_AWAY = 0
    ARM_HOME = 1
    DISARM = 2
    SOS = 3



# Copyright (c) 2022 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Sensor definitions for G90 devices, required when modifying them since writing
a sensor to the device requires values not present on read.
"""
from typing import NamedTuple
from enum import IntEnum


class SensorMatchMode(IntEnum):
    """
    Defines compare (match) mode for the sensor.
    """
    ALL = 0
    ONLY20BITS = 1
    ONLY16BITS = 2


class SensorRwMode(IntEnum):
    """
    Defines read/write mode for the sensor.
    """
    READ = 0
    WRITE = 1
    READ_WRITE = 2


class SensorDefinition(NamedTuple):
    """
    Holds sensor definition data.
    """
    type: int
    subtype: int
    rx: int
    tx: int
    private_data: str
    rwMode: SensorRwMode
    matchMode: SensorMatchMode

    @property
    def reserved_data(self) -> int:
        """
        Sensor's 'reserved_data' field to be written, combined of match
        and RW mode values bitwise.
        """
        return self.matchMode.value << 4 | self.rwMode.value


SENSOR_DEFINITIONS = [
    # Cord Door Sensor
    SensorDefinition(
        type=126,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Glass Sensor
    SensorDefinition(
        type=126,
        subtype=2,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Gas Sensor
    SensorDefinition(
        type=126,
        subtype=3,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Smoke Sensor
    SensorDefinition(
        type=126,
        subtype=4,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord SOS Sensor
    SensorDefinition(
        type=126,
        subtype=5,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Vibration Sensor
    SensorDefinition(
        type=126,
        subtype=6,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Water Sensor
    SensorDefinition(
        type=126,
        subtype=7,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Beam Sensor
    SensorDefinition(
        type=126,
        subtype=9,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord PIR motion sensor
    SensorDefinition(
        type=126,
        subtype=8,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord RFID Sensor
    SensorDefinition(
        type=126,
        subtype=11,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Bell Sensor
    SensorDefinition(
        type=126,
        subtype=12,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Cord Device
    SensorDefinition(
        type=254,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket S07
    SensorDefinition(
        type=128,
        subtype=3,
        rx=0,
        tx=2,
        private_data='060A0600',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket JDQ
    SensorDefinition(
        type=128,
        subtype=0,
        rx=0,
        tx=2,
        private_data='0707070B0B0D0D0E0E00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket Relay (single channel)
    SensorDefinition(
        type=128,
        subtype=1,
        rx=0,
        tx=2,
        private_data='07070700',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket Switch
    SensorDefinition(
        type=128,
        subtype=4,
        rx=0,
        tx=2,
        private_data='050D0500',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket Switch (3 channel)
    SensorDefinition(
        type=128,
        subtype=5,
        rx=0,
        tx=2,
        private_data='070A0E080D060B00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket Switch (4 channel)
    SensorDefinition(
        type=128,
        subtype=6,
        rx=0,
        tx=2,
        private_data='0B0D0E0B0C090A070800',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Rolling curtain
    SensorDefinition(
        type=130,
        subtype=0,
        rx=0,
        tx=7,
        private_data='070B0E0D0C09030100',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Curtain
    SensorDefinition(
        type=130,
        subtype=1,
        rx=0,
        tx=7,
        private_data='0804010200',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Sliding window
    SensorDefinition(
        type=130,
        subtype=1,
        rx=0,
        tx=7,
        private_data='0E0B0E0D0C09030100',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Push Window curtain
    SensorDefinition(
        type=130,
        subtype=2,
        rx=0,
        tx=7,
        private_data='070B0E0D0C09030100',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Air Conditioner
    SensorDefinition(
        type=133,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # TV
    SensorDefinition(
        type=135,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Infrared Transcoder
    SensorDefinition(
        type=145,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Siren SS08
    SensorDefinition(
        type=129,
        subtype=0,
        rx=0,
        tx=2,
        private_data='FEFEF600',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Siren SS07A
    SensorDefinition(
        type=129,
        subtype=4,
        rx=0,
        tx=2,
        private_data='FEFEF600',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Siren SS04
    SensorDefinition(
        type=129,
        subtype=5,
        rx=0,
        tx=2,
        private_data='FCFCCF00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Siren SS02B
    SensorDefinition(
        type=129,
        subtype=1,
        rx=0,
        tx=2,
        private_data='FCFCCF00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Solar-powered siren
    SensorDefinition(
        type=129,
        subtype=2,
        rx=0,
        tx=2,
        private_data='FEFEFD00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Night Light
    SensorDefinition(
        type=138,
        subtype=0,
        rx=0,
        tx=2,
        private_data='060A0600',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Socket_2.4G
    SensorDefinition(
        type=140,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Siren_2.4G
    SensorDefinition(
        type=141,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Switch 2.4G Type 2
    SensorDefinition(
        type=142,
        subtype=2,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Switch 2.4G Type 1
    SensorDefinition(
        type=143,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Switch 2.4G Type 2
    SensorDefinition(
        type=143,
        subtype=2,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Switch 2.4G Type 3
    SensorDefinition(
        type=143,
        subtype=3,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Switch 2.4G Type 4
    SensorDefinition(
        type=143,
        subtype=4,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Curtain 2.4G
    SensorDefinition(
        type=144,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.WRITE,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Door Sensor WDS07
    SensorDefinition(
        type=1,
        subtype=0,
        rx=5,
        tx=0,
        private_data='F100f5017d02f803f90400',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Door Sensor
    SensorDefinition(
        type=1,
        subtype=1,
        rx=2,
        tx=0,
        private_data='0001020400',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Door Sensor WRDS01
    SensorDefinition(
        type=1,
        subtype=3,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Door Sensor
    SensorDefinition(
        type=1,
        subtype=2,
        rx=5,
        tx=0,
        private_data='F100f5017d02f803f904',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Glass Break Sensor BLPS
    SensorDefinition(
        type=2,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Gas Detector WGD01
    SensorDefinition(
        type=3,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Smoke Detector WSD02
    SensorDefinition(
        type=4,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Smoke Detector WSD04
    SensorDefinition(
        type=4,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Panic Button WEB01
    SensorDefinition(
        type=5,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Panic Button WEB03
    SensorDefinition(
        type=5,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Shock Sensor WSS01
    SensorDefinition(
        type=6,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Water Detector LSTC02
    SensorDefinition(
        type=7,
        subtype=1,
        rx=0,
        tx=0,
        private_data='',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Water Detector LSTC01
    SensorDefinition(
        type=7,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor WMS08
    SensorDefinition(
        type=8,
        subtype=3,
        rx=3,
        tx=0,
        private_data='06000301020200',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # PIR motion sensor WMS08B
    SensorDefinition(
        type=8,
        subtype=12,
        rx=3,
        tx=0,
        private_data='060003010202000314',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # PIR motion sensor ODPIR
    SensorDefinition(
        type=8,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor N650
    SensorDefinition(
        type=8,
        subtype=5,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor WPD02
    SensorDefinition(
        type=8,
        subtype=6,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor WCMS02
    SensorDefinition(
        type=8,
        subtype=8,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor CWMS01
    SensorDefinition(
        type=8,
        subtype=9,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor WMS04
    SensorDefinition(
        type=8,
        subtype=11,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor WMS07
    SensorDefinition(
        type=8,
        subtype=2,
        rx=3,
        tx=0,
        private_data='04000801020200',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # PIR motion sensor ODPIR03
    SensorDefinition(
        type=8,
        subtype=4,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor WPD01
    SensorDefinition(
        type=8,
        subtype=7,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # PIR motion sensor PIR Ceiling
    SensorDefinition(
        type=8,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Beams ABT
    SensorDefinition(
        type=9,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Beams ABE
    SensorDefinition(
        type=9,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Beams ABH
    SensorDefinition(
        type=9,
        subtype=2,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Remote RMC08
    SensorDefinition(
        type=10,
        subtype=1,
        rx=4,
        tx=0,
        private_data='0d000b010e02070300',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Remote RMC07
    SensorDefinition(
        type=10,
        subtype=0,
        rx=4,
        tx=0,
        private_data='cf00fc01f3023f0300',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Remote RMC02
    SensorDefinition(
        type=10,
        subtype=4,
        rx=4,
        tx=0,
        private_data='cf00fc01f3023f0300',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Remote
    SensorDefinition(
        type=10,
        subtype=2,
        rx=3,
        tx=0,
        private_data='07000602050300',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # RFID K07
    SensorDefinition(
        type=11,
        subtype=1,
        rx=11,
        tx=0,
        private_data='0e0007010d020b03010402050c060a0709080809060a00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Door Bell WDB
    SensorDefinition(
        type=12,
        subtype=1,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # TouchID Detector
    SensorDefinition(
        type=13,
        subtype=1,
        rx=4,
        tx=0,
        private_data='07000e010d020b0300',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # SOS Watch
    SensorDefinition(
        type=14,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Fingerprint Lock
    SensorDefinition(
        type=15,
        subtype=1,
        rx=0,
        tx=2,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY16BITS
    ),
    # Sub Host SS08S
    SensorDefinition(
        type=16,
        subtype=0,
        rx=0,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ALL
    ),
    # Gas Valve Detector WGD02
    SensorDefinition(
        type=18,
        subtype=0,
        rx=2,
        tx=1,
        private_data='09000801ff0e0e',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
    # Remote 2.4G RMC2.4G
    SensorDefinition(
        type=17,
        subtype=4,
        rx=4,
        tx=0,
        private_data='00',
        rwMode=SensorRwMode.READ,
        matchMode=SensorMatchMode.ONLY20BITS
    ),
]



"""
Tests for G90History class.
"""
import asyncio
from itertools import cycle
from unittest.mock import MagicMock, DEFAULT, ANY
import pytest
from pytest import LogCaptureFixture

from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.history import (
    G90History,
)
from pyg90alarm.const import (
    G90AlertTypes, G90AlertSources,
    G90HistoryStates,
)
from pyg90alarm.exceptions import (
    G90TimeoutError,
)

from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[200,[[50,1,7],'
    b'[3,33,1,1,"Sensor 1",1630147285,""],'
    b'[2,3,0,0,"",1630142877,""],'
    b'[2,5,0,0,"",1630142871,""],'
    b'[2,4,0,0,"",1630142757,""],'
    b'[3,100,1,1,"Sensor 2",1630142297,""],'
    b'[3,1,10,3,"Remote",1734177048,""],'
    b'[1,1,0,0,"",1734175049,""]'
    b']]IEND\0',
])
async def test_history(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving history from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    history = await g90.history(count=7)
    assert len(history) == 7
    assert isinstance(history[0], G90History)
    assert await mock_device.recv_data == [
        b'ISTART[200,200,[200,[1,7]]]IEND\0',
    ]
    assert all(isinstance(h._asdict(), dict) for h in history)
    assert [h._asdict() for h in history] == [
        {
            'datetime': ANY,
            'sensor_idx': None,
            'sensor_name': 'Remote',
            'source': G90AlertSources.REMOTE,
            'state': G90HistoryStates.REMOTE_BUTTON_SOS,
            'type': G90AlertTypes.ALARM,
        },
        {
            'datetime': ANY,
            'sensor_idx': None,
            'sensor_name': None,
            'source': G90AlertSources.DEVICE,
            'state': None,
            'type': G90AlertTypes.HOST_SOS,
        },
        {
            'type': G90AlertTypes.ALARM,
            'source': G90AlertSources.SENSOR,
            'state': G90HistoryStates.DOOR_OPEN,
            'sensor_name': 'Sensor 1',
            'sensor_idx': 33,
            'datetime': ANY,
        },
        {
            'type': G90AlertTypes.STATE_CHANGE,
            'source': G90AlertSources.DEVICE,
            'state': G90HistoryStates.DISARM,
            'sensor_idx': None,
            'sensor_name': None,
            'datetime': ANY,
        },
        {
            'type': G90AlertTypes.STATE_CHANGE,
            'source': G90AlertSources.DEVICE,
            'state': G90HistoryStates.ARM_HOME,
            'sensor_idx': None,
            'sensor_name': None,
            'datetime': ANY,
        },
        {
            'type': G90AlertTypes.STATE_CHANGE,
            'source': G90AlertSources.DEVICE,
            'state': G90HistoryStates.ARM_AWAY,
            'sensor_idx': None,
            'sensor_name': None,
            'datetime': ANY,
        },
        {
            'type': G90AlertTypes.ALARM,
            'source': G90AlertSources.SENSOR,
            'state': G90HistoryStates.DOOR_OPEN,
            'sensor_name': 'Sensor 2',
            'sensor_idx': 100,
            'datetime': ANY,
        },
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[200,[[3,1,3],'
    # Wrong state
    b'[3,33,7,200,"Sensor 1",1630147285,""],'
    # Wrong source
    b'[2,33,254,1,"Sensor 1",1630147285,""],'
    # Wrong type
    b'[254,33,1,1,"Sensor 1",1630147285,""]'
    b']]IEND\0',
])
async def test_history_parsing_error(mock_device: DeviceMock) -> None:
    """
    Tests for processing history from the device, when the parsing error
    occurs.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    history = await g90.history(count=5)
    assert len(history) == 3
    assert isinstance(history[0], G90History)
    assert isinstance(history[0]._asdict(), dict)
    # Wrong entry element should result in corresponding key having 'None'
    # value
    assert history[0]._asdict()['state'] is None
    assert history[1]._asdict()['source'] is None
    assert history[2]._asdict()['type'] is None


@pytest.mark.g90device(sent_data=[
    # Host info
    b'ISTART[206,'
    b'["DUMMYGUID","DUMMYPRODUCT",'
    b'"1.2","1.1","206","206",3,3,0,2,"4242",50,100]]IEND\0',
    # Simulate empty history initially
    b'ISTART[200,[[0,0,0]]]IEND\0',
    # The history records will be used to remember the timestamp of most recent
    # one
    b'ISTART[200,[[1,1,1],'
    b'[2,5,0,0,"",1630142871,""]'
    b']]IEND\0',
    # The records will be used to simulate the device alerts, but only for
    # those newer that one above
    b'ISTART[200,[[3,1,3],'
    b'[3,33,1,1,"Sensor 1",1630147285,""],'
    b'[2,3,0,0,"",1630142877,""],'
    b'[2,5,0,0,"",1630142871,""]'
    b']]IEND\0',
    # Simulated list of devices, will be used by alarm callback
    b'ISTART[102,'
    b'[[2,1,2],'
    b'["Sensor 1",33,0,138,0,0,33,0,0,17,1,0,""],'
    b'["Sensor 2",100,0,138,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    # Alert configuration, used by sensor activity callback invoked when
    # handling alarm
    b'ISTART[117,[256]]IEND\0',
])
async def test_simulate_alerts_from_history(mock_device: DeviceMock) -> None:
    """
    Tests for simulating device alerts from the history.
    """
    # Callback handlers for alarm and arm/disarm, just setting their
    # corresponding future when called
    future_alarm = asyncio.get_running_loop().create_future()
    future_armdisarm = asyncio.get_running_loop().create_future()
    alarm_cb = MagicMock()
    alarm_cb.side_effect = lambda *args: future_alarm.set_result(True)
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future_armdisarm.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    # Call the method to store device GUID, so that its validation in
    # `G90DeviceNotifications._handle_alert()` is involved
    await g90.get_host_info()
    g90.alarm_callback = alarm_cb
    g90.armdisarm_callback = armdisarm_cb
    # Simulate device timeout exception every 2nd call to `G90Alarm.history()`
    # method - the processing should still result in callbacks invoked
    g90.history = MagicMock(wraps=g90.history)  # type: ignore[method-assign]
    g90.history.side_effect = cycle([G90TimeoutError, DEFAULT])
    # Simulate device notifications from the history data above, small interval
    # is set to shorten the test run time
    await g90.start_simulating_alerts_from_history(interval=0.1)
    # Both callbacks should be called, wait for that - the timeout should be
    # sufficient for extra iterations in the method under test, to accommodate
    # the simulated exceptions above
    await asyncio.wait([future_alarm, future_armdisarm], timeout=0.5)
    # Stop simulating the alert from history
    await g90.stop_simulating_alerts_from_history()

    sensors = await g90.sensors
    # Ensure callbacks have been called and with expected arguments
    alarm_cb.assert_called_once_with(33, 'Sensor 1', None)
    armdisarm_cb.assert_called_once_with(3)
    assert sensors[0].occupancy is True


async def test_simulate_alerts_from_history_exception(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Tests for simulating device alerts from the history, when an exception is
    raised when interacting with the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    # Simulate a generic error fetching history entries
    g90.history = MagicMock()  # type: ignore[method-assign]
    simulated_error = Exception('dummy error')
    g90.history.side_effect = simulated_error
    caplog.set_level('WARNING')
    # Start simulating alerts from history
    await g90.start_simulating_alerts_from_history()
    # Allow task to settle
    await asyncio.sleep(0.1)
    # Verify the task is no longer running and resulted in particular exception
    task = g90._alert_simulation_task  # pylint:disable=protected-access
    assert task is not None
    assert task.exception() == simulated_error
    assert task.done()
    # Stop simulating the alert from history
    await g90.stop_simulating_alerts_from_history()
    # Verify the error logged
    assert ''.join(caplog.messages).startswith(
        'Exception simulating device alerts from history'
    )



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Defines notification protocol for `NotificationBase` class.
"""

from ..const import (
    G90ArmDisarmTypes,
    G90RemoteButtonStates,
)


class G90NotificationProtocol:
    """
    Protocol for notification handling.
    """
    async def on_armdisarm(self, state: G90ArmDisarmTypes) -> None:
        """
        Invoked when device is armed or disarmed.

        :param state: State of the device
        """

    async def on_sensor_activity(self, idx: int, name: str) -> None:
        """
        Invoked on sensor activity.

        :param idx: Index of the sensor.
        :param name: Name of the sensor.
        """

    async def on_door_open_when_arming(
        self, event_id: int, zone_name: str
    ) -> None:
        """
        Invoked when door open is detected when panel is armed.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the sensor that reports door open.
        """

    async def on_door_open_close(
        self, event_id: int, zone_name: str, is_open: bool
    ) -> None:
        """
        Invoked when door sensor reports it opened or closed.

        :param event_id: Index of the sensor reporting the event.
        :param zone_name: Name of the sensor that reports door open/close.
        :param is_open: Indicates if the door is open.
        """

    async def on_low_battery(self, event_id: int, zone_name: str) -> None:
        """
        Invoked when a sensor reports it is low on battery.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the sensor that reports low battery.
        """

    async def on_alarm(
        self, event_id: int, zone_name: str, is_tampered: bool
    ) -> None:
        """
        Invoked when device triggers the alarm.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the zone that triggered the alarm.
        """

    async def on_remote_button_press(
        self, event_id: int, zone_name: str, button: G90RemoteButtonStates
    ) -> None:
        """
        Invoked when a remote button is pressed.

        Please note there will only be call to the method w/o invoking
        :meth:`G90DeviceNotifications.on_sensor_activity`.

        :param event_id: Index of the sensor associated with the remote.
        :param zone_name: Name of the sensor that reports remote button press.
        :param button: The button pressed on the remote
        """

    async def on_sos(
        self, event_id: int, zone_name: str, is_host_sos: bool
    ) -> None:
        """
        Invoked when SOS is triggered.

        Please note that the panel might not set its status to alarm
        internally, so that :meth:`G90DeviceNotifications` might need an
        explicit call in the derived class to simulate that.

        :param event_id: Index of the sensor.
        :param zone_name: Name of the sensor that reports SOS.
        :param is_host_sos: Indicates if the SOS is host-initiated.
        """



"""
Tests for G90DeviceNotifications class
"""
import asyncio
import re
from unittest.mock import MagicMock
import pytest
from pytest import LogCaptureFixture

from pyg90alarm.local.notifications import (
    G90LocalNotifications,
)
from pyg90alarm.alarm import G90Alarm
from pyg90alarm.notifications.protocol import G90NotificationProtocol

from .device_mock import DeviceMock
from .conftest import data_receive_awaitable


@pytest.mark.g90device(notification_data=[
    b'\xdeadbeef\0',
])
async def test_device_notification_invalid_utf_data(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that wrong UTF encoded data in device notification is handled
    correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('ERROR')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert ''.join(caplog.messages) == (
        "Unable to decode device message from UTF-8"
    )


@pytest.mark.g90device(notification_data=[
    b'[170]\0',
])
async def test_device_notification_missing_header(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that missing header in device notification is handled correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('ERROR')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert re.match(
        r"Device message '\[170\]' is malformed: .+ missing 1 required"
        " positional argument: 'data'",
        ''.join(caplog.messages)
    )


@pytest.mark.g90device(notification_data=[
    b'[170,[1,[1]]\0',
])
async def test_device_notification_malformed_message(
    mock_device: DeviceMock,
    caplog: LogCaptureFixture
) -> None:
    """
    Verifies that malformed message in device notification is handled
    correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('ERROR')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert (
        "Unable to parse device message '[170,[1,[1]]' as JSON:"
        in ''.join(caplog.messages)
    )


@pytest.mark.g90device(notification_data=[
    b'[170,[1,[1]]]',
])
async def test_device_notification_missing_end_marker(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that missing end marker in device notification is handled
    correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('ERROR')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert ''.join(caplog.messages) == 'Missing end marker in data'


@pytest.mark.g90device(notification_data=[
    b'[170,[1]]\0',
])
async def test_wrong_device_notification_format(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that wrong device notification format is handled correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('ERROR')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert re.match(
        'Bad notification received:'
        " .+ missing 1 required positional argument: 'data'",
        ''.join(caplog.messages)
    )


@pytest.mark.g90device(notification_data=[
    b'[208,[]]\0',
])
async def test_wrong_device_alert_format(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that wrong device alert format is handled correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('ERROR')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert re.match(
        'Bad alert received:'
        " .+ missing 9 required positional arguments: 'type',"
        " 'event_id', 'source', 'state', 'zone_name', 'device_id',"
        " 'unix_time', 'resv4', and 'other'",
        ''.join(caplog.messages)
    )


@pytest.mark.g90device(notification_data=[
    b'[170,[999,[1]]]\0',
])
async def test_unknown_device_notification(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that unknown device notification is handled correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('WARNING')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert re.match(
        'Unknown notification received:'
        r' kind 999, data \[1\]',
        ''.join(caplog.messages)
    )


@pytest.mark.g90device(notification_data=[
    b'[208,[999,100,1,1,"Hall","DUMMYGUID",'
    b'1631545189,0,[""]]]\0',
])
async def test_unknown_device_alert(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that unknown device alert is handled correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host=mock_device.host,
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('WARNING')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert re.match(
        'Unknown alert received: type 999,'
        r' data G90DeviceAlert\(type=999, event_id=100, source=1,'
        r" state=1, zone_name='Hall', device_id='DUMMYGUID',"
        r" unix_time=1631545189, resv4=0, other=\[''\]\)",
        ''.join(caplog.messages)
    )


@pytest.mark.g90device(notification_data=[
    b'[208,[999,100,1,1,"Hall","DUMMYGUID",'
    b'1631545189,0,[""]]]\0',
])
async def test_wrong_host(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that unknown device alert is handled correctly.
    """
    notifications = G90LocalNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        host='1.2.3.4',
        port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    notifications.handle_alert = (  # type: ignore[method-assign]
        MagicMock()
    )
    notifications.handle_notification = (  # type: ignore[method-assign]
        MagicMock()
    )

    caplog.set_level('WARNING')
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert ''.join(caplog.messages) == (
        "Received notification/alert from wrong host '127.0.0.1'"
        ", expected from '1.2.3.4'"
    )
    notifications.handle_alert.assert_not_called()
    notifications.handle_notification.assert_not_called()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[206,'
        b'["","DUMMYPRODUCT",'
        b'"1.2","1.1","206","206",3,3,0,2,"4242",50,100]]IEND\0',
    ],
)
async def test_empty_device_guid(mock_device: DeviceMock) -> None:
    """
    Verifies that alert from device with empty GUID is ignored.
    """
    g90 = G90Alarm(
        host=mock_device.host, port=mock_device.port
    )
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    # The command will fetch the host info and store the GIUD
    await g90.get_host_info()
    await g90.close_notifications()
    # pylint: disable=protected-access
    assert g90._notifications is not None
    assert g90._notifications.device_id is None


@pytest.mark.g90device(notification_data=[
    b'[208,[2,4,0,0,"","DIFFERENTGUID",1630876128,0,[""]]]\0'
])
async def test_wrong_device_guid(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that alert from device with different GUID is ignored.
    """
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )
    future = data_receive_awaitable(notifications)

    caplog.set_level('WARNING')
    notifications.device_id = 'DUMMYGUID'
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    assert ''.join(caplog.messages) == (
        "Received alert from wrong device: expected 'DUMMYGUID'"
        ", got 'DIFFERENTGUID'"
    )
    # Verify the associated callback was not called
    mock.on_armdisarm.assert_not_called()


@pytest.mark.g90device(notification_data=[
    b'[170,[5,[100,"Hall"]]]\0',
])
async def test_sensor_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that sensor notification callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_sensor_activity.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_sensor_activity.assert_called_once_with(100, 'Hall')
    # Verify last packet from device got tracked
    assert notifications.last_device_packet_time is not None
    # Contrary, the local notification never send anything upstream - verify
    # that
    assert notifications.last_upstream_packet_time is None


@pytest.mark.g90device(notification_data=[
    b'[170,[1,[1]]]\0',
])
async def test_armdisarm_notification_callback(
    mock_device: DeviceMock
) -> None:
    """
    Verifies that arm/disarm notification callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_armdisarm.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_armdisarm.assert_called_once_with(1)


@pytest.mark.g90device(notification_data=[
    b'[208,[2,4,0,0,"","DUMMYGUID",1630876128,0,[""]]]\0',
])
async def test_armdisarm_alert_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that arm/disarm alert callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_armdisarm.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_armdisarm.assert_called_once_with(1)


@pytest.mark.g90device(notification_data=[
    b'[208,[4,100,1,1,"Hall","DUMMYGUID",1631545189,0,[""]]]\0',
])
async def test_door_open_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that door open callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_door_open_close.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_door_open_close.assert_called_once_with(100, 'Hall', True)


@pytest.mark.g90device(notification_data=[
    b'[208,[4,100,1,0,"Hall","DUMMYGUID",1631545189,0,[""]]]\0',
])
async def test_door_close_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that door close callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_door_open_close.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_door_open_close.assert_called_once_with(
        100, 'Hall', False
    )


@pytest.mark.g90device(notification_data=[
    b'[208,[4,111,12,0,"Doorbell","DUMMYGUID",1655745021,0,[""]]]\0',
])
async def test_doorbell_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that doorbell callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_door_open_close.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_door_open_close.assert_called_once_with(
        111, 'Doorbell', True
    )


@pytest.mark.g90device(notification_data=[
    b'[208,[3,11,1,1,"Hall","DUMMYGUID",1630876128,0,[""]]]\0',
])
async def test_alarm_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that alarm callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_alarm.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_alarm.assert_called_once_with(11, 'Hall', False)


@pytest.mark.g90device(notification_data=[
    b'[208,[3,11,1,3,"Hall","DUMMYGUID",1630876128,0,[""]]]\0',
])
async def test_tamper_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that alarm callback is handled correctly when a sensor is
    tampered.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_alarm.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_alarm.assert_called_once_with(11, 'Hall', True)


@pytest.mark.g90device(notification_data=[
    # Host SOS
    b'[208,[1,1,0,0,"","DUMMYGUID",1734175050,0,[""]]]\0',
    # Remote SOS
    b'[208,[3,1,10,3,"Remote","DUMMYGUID",1734177048,0,[""]]]\0',
])
async def test_sos_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that remote SOS callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_sos.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()

    # Host SOS
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    mock.on_sos.assert_called_with(1, 'Host SOS', True)

    # Remote SOS
    mock.on_sos.reset_mock()
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    mock.on_sos.assert_called_with(1, 'Remote', False)

    await notifications.close()


@pytest.mark.g90device(notification_data=[
    b'[208,[4,26,1,4,"Hall","DUMMYGUID",1719223959,0,[""]]]\0'
])
async def test_low_battery_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that low battery callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_low_battery.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_low_battery.assert_called_once_with(26, 'Hall')


@pytest.mark.g90device(notification_data=[
    b'[170,[6,[21,"Hall"]]]\0'
])
async def test_door_open_when_arming_callback(mock_device: DeviceMock) -> None:
    """
    Verifies that door open when arming callback is handled correctly.
    """
    future = asyncio.get_running_loop().create_future()
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90LocalNotifications(
        protocol_factory=lambda: mock,
        host=mock_device.host, port=mock_device.port,
        local_host=mock_device.notification_host,
        local_port=mock_device.notification_port
    )

    mock.on_door_open_when_arming.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_door_open_when_arming.assert_called_once_with(21, 'Hall')



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Cloud communication implementation for G90 alarm systems.

This module provides the necessary components to interact with G90 alarm
systems through their cloud protocol (referred to as version 1.1).
"""
from .notifications import G90CloudNotifications  # noqa: F401

__all__ = [
    'G90CloudNotifications',
]



"""
Tests for G90CloudNotifications class.
"""
import asyncio
from unittest.mock import MagicMock, patch
from dataclasses import dataclass
from struct import pack

import pytest
from freezegun import freeze_time

from pyg90alarm.notifications.protocol import G90NotificationProtocol
from pyg90alarm.cloud import G90CloudNotifications
from pyg90alarm.cloud.protocol import (
    G90CloudMessage
)
from pyg90alarm.cloud.messages import (
    cloud_message
)
from pyg90alarm.cloud.const import G90CloudDirection, G90CloudCommand

from .device_mock import DeviceMock
from .conftest import data_receive_awaitable


@patch('pyg90alarm.cloud.messages.CLOUD_MESSAGE_CLASSES', [])
async def test_cloud_duplicate_message_defintion() -> None:
    """
    Tests that duplicate cloud message definitions raise a ValueError.
    """
    @dataclass
    @cloud_message
    class _TestMessage1(G90CloudMessage):
        _command = G90CloudCommand.HELLO
        _source = G90CloudDirection.DEVICE
        _destination = G90CloudDirection.CLOUD

    with pytest.raises(ValueError):
        @dataclass
        @cloud_message
        class _TestMessage2(G90CloudMessage):
            _command = G90CloudCommand.HELLO
            _source = G90CloudDirection.DEVICE
            _destination = G90CloudDirection.CLOUD


@pytest.mark.g90device(cloud_notification_data=[
    b'\x01\x10\x00\x00\x08',
])
async def test_cloud_short_header(
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of cloud packets with headers that are too short.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await notifications.close()
    assert await mock_device.cloud_recv_data == []


@pytest.mark.g90device(cloud_notification_data=[
    b'\x01\x10\x00\x00\x09\x00\x00\x00',
])
async def test_cloud_short_packet(
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of cloud packets that are shorter than expected length.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert await mock_device.cloud_recv_data == []


@pytest.mark.g90device(cloud_notification_data=[
    b'\x01\x10\x00\x00\x08\x00\x00\x00',
])
async def test_cloud_ping(
    mock_device: DeviceMock
) -> None:
    """
    Tests the correct processing and response to cloud ping messages.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert (
        b'\x01\x10\x00\x00\x08\x00\x00\x00'
    ) == bytes().join(await mock_device.cloud_recv_data)


@pytest.mark.g90device(cloud_notification_data=[
    b'\x01\x10\x00\x20\x48\x00\x00\x00\x01\x00\x00\x00'
    b'\x47\x41\x30\x30\x30\x30\x30\x41\x30\x30\x30\x30'
    b'\x30\x30\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x02\x00\x00\x00\x00\x70\x00\x00\x32\x30\x37\x00'
    b'\x58\xba\x00\x20\x30\x00\x00\x00\x00\x00\x00\x00'
    b'\b07\x00\x00\x00\x1e\x00\x00\x00\x1e\x00'
])
async def test_cloud_hello(
    mock_device: DeviceMock
) -> None:
    """
    Tests the proper handling and response to cloud hello messages from
    devices.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert (
        b'\x41\x20\x00\x10\x0d\x00\x00\x00\x01\x00\x01\x00\x01'
        b'\x01\x20\x00\x10\x0d\x00\x00\x00\x01\x00\x02\x00\x1f'
        b'\x63\x20\x00\x10\x10\x00\x00\x00\x01\x00\x03\x00'
        + pack('<i', mock_device.cloud_port)
    ) == bytes().join(await mock_device.cloud_recv_data)
    # Verify the last packet from the device got tracked
    assert notifications.last_device_packet_time is not None


@pytest.mark.g90device(cloud_notification_data=[
    b'\x01\x10\x00\x20\x48\x00\x00\x00\x11\x00\x00\x00'
    b'\x47\x41\x30\x30\x30\x30\x30\x41\x30\x30\x30\x30'
    b'\x30\x30\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x02\x00\x00\x00\x00\x70\x00\x00\x32\x30\x37\x00'
    b'\x58\xba\x00\x20\x30\x00\x00\x00\x00\x00\x00\x00'
    b'\b07\x00\x00\x00\x1e\x00\x00\x00\x1e\x00'
])
async def test_cloud_hello_wrong_version(
    mock_device: DeviceMock
) -> None:
    """
    Tests rejection of cloud hello messages with an incorrect version number.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert await mock_device.cloud_recv_data == []


@freeze_time("2025-01-01 00:00:00")
@pytest.mark.g90device(cloud_notification_data=[
    b'\x01\x30\x00\x20\x3c\x00\x00\x00\x01\x00\x00\x00\x47\x41\x30\x30'
    b'\x30\x30\x30\x41\x30\x30\x30\x30\x30\x30\x30\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x01\x00\x00\x00\x00\x70\x00\x00\x32\x30\x37\x00'
    b'\x05\x05\x05\x05\x30\x00\x06\x06\x07\x07\x07\x07'
])
async def test_cloud_hello_discovery(
    mock_device: DeviceMock
) -> None:
    """
    Tests that devices respond correctly to discovery hello messages from the
    cloud.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert (
        b'\x01\xd0\x00\x10\x2c\x00\x00\x00\x01\x00\x00\x00\x34\x37\x2e\x38'
        b'\x38\x2e\x37\x2e\x36\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x2e\x16\x00\x00\x80\x85\x74\x67'
    ) == bytes().join(await mock_device.cloud_recv_data)


@pytest.mark.g90device(
    cloud_notification_data=[
        b'\x22\x10\x00\x20\x25\x00\x00\x00\x01\x00\x00\x00\x5b\x31\x37\x30'
        b'\x2c\x5b\x35\x2c\x5b\x31\x30\x31\x2c\x22\x43\x6f\x72\x64\x20\x31'
        b'\x22\x5d\x5d\x5d\x00',
    ],
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Cord 1",101,0,126,1,0,32,0,5,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ]
)
async def test_cloud_notification(
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of cloud notifications about sensor activity.
    """
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90CloudNotifications(
        protocol_factory=lambda: mock,
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = asyncio.get_running_loop().create_future()
    mock.on_sensor_activity.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert await mock_device.cloud_recv_data == []
    mock.on_sensor_activity.assert_called_once_with(101, 'Cord 1')


@pytest.mark.g90device(cloud_notification_data=[
    # Hello message, to verify device ID is stored properly
    b'\x01\x10\x00\x20\x48\x00\x00\x00\x01\x00\x00\x00'
    b'\x47\x41\x30\x30\x30\x30\x30\x41\x30\x30\x30\x30'
    b'\x30\x30\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x02\x00\x00\x00\x00\x70\x00\x00\x32\x30\x37\x00'
    b'\x58\xba\x00\x20\x30\x00\x00\x00\x00\x00\x00\x00'
    b'\b07\x00\x00\x00\x1e\x00\x00\x00\x1e\x00',
    # Disarm message
    b'\x21\x10\x00\x20\x78\x00\x00\x00\x01\x00\x00\x00\x02\x03\x37\x30'
    b'\x2c\x5b\x31\x2c\x5b\x33\x5d\x5d\x5d\x00\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x20\xa2\x94\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00'
])
async def test_cloud_status_change_disarm(
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of disarm status changes from the cloud.
    """
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90CloudNotifications(
        protocol_factory=lambda: mock,
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = asyncio.get_running_loop().create_future()
    mock.on_armdisarm.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    mock.on_armdisarm.assert_called_once_with(3)


@pytest.mark.parametrize("_title,expected_on_alarm_args", [
    pytest.param("tamper", (32, 'Cord 1', True), marks=pytest.mark.g90device(
        cloud_notification_data=[
            # Tamper alarm
            b'\x21\x10\x00\x20\x78\x00\x00\x00\x01\x00\x00\x00\x03\x20\x01\x03'
            b'\x43\x6f\x72\x64\x20\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x0f\xdd\xa9\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Cord 1",32,0,126,1,0,32,0,5,16,1,0,""]]]IEND\0',
            b'ISTART[117,[256]]IEND\0',
        ])
    ),
    pytest.param("gas", (32, 'Cord 1', False), marks=pytest.mark.g90device(
        cloud_notification_data=[
            # Gas sensor alarm
            b'\x21\x10\x00\x20\x78\x00\x00\x00\x01\x00\x00\x00\x03\x20\x03\xfe'
            b'\x43\x6f\x72\x64\x20\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\xd6\x85\xc4\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Cord 1",32,0,126,1,0,32,0,5,16,1,0,""]]]IEND\0',
            b'ISTART[117,[256]]IEND\0',
        ])
    ),
    pytest.param("regular", (32, 'Cord 1', False), marks=pytest.mark.g90device(
        cloud_notification_data=[
            # Regular alarm
            b'\x21\x10\x00\x20\x78\x00\x00\x00\x01\x00\x00\x00\x03\x20\x08\x00'
            b'\x43\x6f\x72\x64\x20\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x08\x28\xa2\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Cord 1",32,0,126,1,0,32,0,5,16,1,0,""]]]IEND\0',
            b'ISTART[117,[256]]IEND\0',
        ])
    ),
])
async def test_cloud_status_change_alarm_sensor(
    _title: str,
    expected_on_alarm_args: tuple[int, str, bool],
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of alarm status changes from sensors.
    """
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90CloudNotifications(
        protocol_factory=lambda: mock,
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = asyncio.get_running_loop().create_future()
    mock.on_alarm.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert await mock_device.cloud_recv_data == []
    mock.on_alarm.assert_called_once_with(*expected_on_alarm_args)


@pytest.mark.g90device(
    cloud_notification_data=[
        b'\x21\x10\x00\x20\x78\x00\x00\x00\x01\x00\x00\x00\x04\x20\x01\x01'
        b'\x43\x6f\x72\x64\x20\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\xb1\x30\xa2\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00'
    ],
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Cord 1",32,0,126,1,0,32,0,5,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ]
)
async def test_cloud_status_change_sensor_activity(
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of sensor activity status changes from the cloud.
    """
    mock = MagicMock(spec=G90NotificationProtocol)
    notifications = G90CloudNotifications(
        protocol_factory=lambda: mock,
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = asyncio.get_running_loop().create_future()
    mock.on_door_open_close.side_effect = (
        lambda *args: future.set_result(True)
    )
    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()
    assert await mock_device.cloud_recv_data == []
    mock.on_door_open_close.assert_called_once_with(32, 'Cord 1', True)


@pytest.mark.g90device(
    cloud_notification_data=[
        b'\x01\x00\x00\x20\x3b\x00\x00\x00\x01\x00\x06\x00\x00\x00\x00\x00'
        b'\x06\x00\x00\x00\x6a\x00\x6a\x00\x5b\x31\x30\x36\x2c\x5b\x39\x30'
        b'\x30\x2c\x30\x2c\x30\x2c\x31\x2c\x32\x2c\x32\x2c\x36\x30\x2c\x32'
        b'\x2c\x30\x2c\x31\x38\x30\x2c\x32\x5d\x5d\x00'
    ]
)
async def test_cloud_command(
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of cloud commands sent to the device.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    # Cloud command shouldn't send any response
    assert await mock_device.cloud_recv_data == []


@pytest.mark.parametrize("_title", [
    pytest.param("simple", marks=pytest.mark.g90device(
        cloud_notification_data=[
            b'\x01\x10\x00\x20\x48\x00\x00\x00\x01\x00\x00\x00'
            b'\x47\x41\x30\x30\x30\x30\x30\x41\x30\x30\x30\x30'
            b'\x30\x30\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x02\x00\x00\x00\x00\x70\x00\x00\x32\x30\x37\x00'
            b'\x58\xba\x00\x20\x30\x00\x00\x00\x00\x00\x00\x00'
            b'\b07\x00\x00\x00\x1e\x00\x00\x00\x1e\x00'],
        # The data isn't parsed by the package currently, rather serves as a
        # documented message example for the future changes
        cloud_upstream_data=[
            b'\x29\x50\x00\x10\x17\x00\x00\x00\x01\x00\x06\x00\x02\x00\x64\x00'
            b'\x64\x00\x05\x00\x00\x00\x00',
        ]
    )),
    pytest.param("paginated", marks=pytest.mark.g90device(
        cloud_notification_data=[
            b'\x01\x10\x00\x20\x48\x00\x00\x00\x01\x00\x00\x00'
            b'\x47\x41\x30\x30\x30\x30\x30\x41\x30\x30\x30\x30'
            b'\x30\x30\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x02\x00\x00\x00\x00\x70\x00\x00\x32\x30\x37\x00'
            b'\x58\xba\x00\x20\x30\x00\x00\x00\x00\x00\x00\x00'
            b'\b07\x00\x00\x00\x1e\x00\x00\x00\x1e\x00'],
        # See above
        cloud_upstream_data=[
            b'\x29\x50\x00\x10\x23\x00\x00\x00\x01\x00\x05\x00\x02\x00\xc8\x00'
            b'\xc8\x00\x05\x00\x00\x00\x5b\x32\x30\x30\x2c\x5b\x31\x2c\x31\x30'
            b'\x5d\x5d\x00',
        ]
    )),
])
async def test_upstream_cloud_hello(
    _title: str,
    mock_device: DeviceMock
) -> None:
    """
    Tests handling of upstream cloud commands sent to the device.
    """
    notifications = G90CloudNotifications(
        protocol_factory=lambda: MagicMock(spec=G90NotificationProtocol),
        local_host=mock_device.cloud_host,
        local_port=mock_device.cloud_port,
        upstream_host=mock_device.cloud_upstream_host,
        upstream_port=mock_device.cloud_upstream_port
    )
    future = data_receive_awaitable(notifications)

    await notifications.listen()
    await mock_device.send_next_cloud_packet()
    await asyncio.wait([future], timeout=0.1)
    await notifications.close()

    # Verify the data from the device is sent to the upstream service umodified
    assert await mock_device.cloud_upstream_recv_data == mock_device.cloud_data
    # Verify the data from simulated upstream service is sent back to the
    # device unmodified
    assert await mock_device.cloud_recv_data == mock_device.cloud_upstream_data
    # Verify both device and upstream packet times are set
    assert notifications.last_device_packet_time is not None
    assert notifications.last_upstream_packet_time is not None



'''
Tests for G90PaginatedCommand class
'''
import pytest

from pyg90alarm.local.paginated_cmd import (
    G90PaginatedCommand,
)
from pyg90alarm.exceptions import G90Error
from pyg90alarm.const import G90Commands

from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,[[]]]IEND\0',
])
async def test_missing_pagination_header(mock_device: DeviceMock) -> None:
    '''
    Verifies that missing pagination header raises an exception.
    '''
    g90 = G90PaginatedCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETSENSORLIST,
        start=1, end=1
    )

    with pytest.raises(
        G90Error,
        match=(
            r"Wrong pagination data \[\] -"
            ' .+ missing 3 required positional arguments:'
            " 'total', 'start', and 'nelems'"
        )
    ):
        await g90.process()

    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,1]]]IEND\0'
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,[[1,1,1]]]IEND\0',
])
async def test_no_paginated_data(mock_device: DeviceMock) -> None:
    '''
    Verifies that no paginated data raises an exception.
    '''
    g90 = G90PaginatedCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETSENSORLIST,
        start=1, end=1
    )

    with pytest.raises(
        G90Error,
        match=(
            'Truncated data provided in paginated response -'
            ' expected 1 entities as per response, received 0'
        )
    ):
        await g90.process()
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,1]]]IEND\0'
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,[[1,1,1],[""]]]IEND\0',
])
async def test_partial_paginated_data(mock_device: DeviceMock) -> None:
    '''
    Verifies that paginated data shorter than requested is properly handled.
    '''
    g90 = G90PaginatedCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETSENSORLIST,
        start=1, end=2
    )

    await g90.process()
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,2]]]IEND\0'
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,[[2,1,2],[""],[""]]]IEND\0',
])
async def test_extra_paginated_data(mock_device: DeviceMock) -> None:
    '''
    Verifies that extra paginated data raises an exception.
    '''
    g90 = G90PaginatedCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETSENSORLIST,
        start=1, end=1
    )

    with pytest.raises(
        G90Error,
        match=(
            'Extra data provided in paginated response -'
            ' expected 1 entities as per request, received 2'
        )
    ):
        await g90.process()
    assert await mock_device.recv_data == [
        b'ISTART[102,102,[102,[1,1]]]IEND\0'
    ]



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides interface to sensors of G90 alarm panel.
"""
from __future__ import annotations
import logging
from dataclasses import dataclass, asdict, astuple
from typing import (
    Any, Optional, TYPE_CHECKING, Dict
)

from enum import IntEnum, IntFlag
from ..definitions.sensors import SENSOR_DEFINITIONS, SensorDefinition
from ..const import G90Commands
from .base_entity import G90BaseEntity
if TYPE_CHECKING:
    from ..alarm import (
        G90Alarm, SensorStateCallback, SensorLowBatteryCallback,
        SensorDoorOpenWhenArmingCallback, SensorTamperCallback,
    )


@dataclass
class G90SensorCommonData:  # pylint:disable=too-many-instance-attributes
    """
    Common protocol fields across read and write operations.

    :meta private:
    """
    parent_name: str
    index: int
    room_id: int
    type_id: int
    subtype: int
    timeout: int
    user_flag_data: int
    baudrate: int
    protocol_id: int
    reserved_data: int
    node_count: int


@dataclass
class G90SensorIncomingData(G90SensorCommonData):
    """
    Incoming (read operation) protocol fields.

    :meta private:
    """
    mask: int
    private_data: str


@dataclass
class G90SensorOutgoingData(G90SensorCommonData):
    """
    Outgoing (write operation) protocol fields.

    :meta private:
    """
    rx: int  # pylint:disable=invalid-name
    tx: int  # pylint:disable=invalid-name
    private_data: str


class G90SensorReservedFlags(IntFlag):
    """
    Reserved flags of the sensor.
    """
    NONE = 0
    CAN_READ = 16
    CAN_READ_EXT = 32
    CAN_WRITE = 1


class G90SensorUserFlags(IntFlag):
    """
    User flags of the sensor.
    """
    NONE = 0
    ENABLED = 1
    ARM_DELAY = 2
    DETECT_DOOR = 4
    DOOR_CHIME = 8
    INDEPENDENT_ZONE = 16
    ALERT_WHEN_AWAY_AND_HOME = 32
    ALERT_WHEN_AWAY = 64
    SUPPORTS_UPDATING_SUBTYPE = 512     # Only relevant for cord sensors
    # Flags that can be set by the user
    USER_SETTABLE = (
        ENABLED
        | ARM_DELAY
        | DETECT_DOOR
        | DOOR_CHIME
        | INDEPENDENT_ZONE
        | ALERT_WHEN_AWAY_AND_HOME
        | ALERT_WHEN_AWAY
    )


class G90SensorProtocols(IntEnum):
    """
    Protocol types for the sensors.
    """
    RF_1527 = 0
    RF_2262 = 1
    RF_PRIVATE = 2
    RF_SLIDER = 3
    CORD = 5
    WIFI = 4
    USB = 6


class G90SensorTypes(IntEnum):
    """
    Sensor types.
    """
    DOOR = 1
    GLASS = 2
    GAS = 3
    SMOKE = 4
    SOS = 5
    VIB = 6
    WATER = 7
    INFRARED = 8
    IN_BEAM = 9
    REMOTE = 10
    RFID = 11
    DOORBELL = 12
    BUTTONID = 13
    WATCH = 14
    FINGER_LOCK = 15
    SUBHOST = 16
    REMOTE_2_4G = 17
    CORD_SENSOR = 126
    SOCKET = 128
    SIREN = 129
    CURTAIN = 130
    SLIDINGWIN = 131
    AIRCON = 136
    TV = 137
    NIGHTLIGHT = 138
    SOCKET_2_4G = 140
    SIREN_2_4G = 141
    SWITCH_2_4G = 142
    TOUCH_SWITCH_2_4G = 143
    CURTAIN_2_4G = 144
    CORD_DEV = 254
    UNKNOWN = 255


_LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-public-methods
class G90Sensor(G90BaseEntity):  # pylint:disable=too-many-instance-attributes
    """
    Interacts with sensor on G90 alarm panel.

    :param args: Pass-through positional arguments for for interpreting
     protocol fields
    :param parent: Instance of :class:`.G90Alarm` the sensor is associated
     with
    :type parent: :class:`.G90Alarm`
    :param int subindex: Index of the sensor within multi-channel devices
     (those having multiple nodes)
    :param int proto_idx: Index of the sensor within list of sensors as
     retrieved from the alarm panel
    :param kwargs: Pass-through keyword arguments for for interpreting protocol
     fields
    """
    def __init__(
        self, *args: Any, parent: G90Alarm, subindex: int, proto_idx: int,
        **kwargs: Any
    ) -> None:
        self._protocol_incoming_data_kls = G90SensorIncomingData
        self._protocol_outgoing_data_kls = G90SensorOutgoingData
        self._protocol_data = self._protocol_incoming_data_kls(*args, **kwargs)
        self._parent = parent
        self._subindex = subindex
        self._occupancy = False
        self._state_callback: Optional[SensorStateCallback] = None
        self._low_battery_callback: Optional[SensorLowBatteryCallback] = None
        self._low_battery = False
        self._tampered = False
        self._door_open_when_arming_callback: Optional[
            SensorDoorOpenWhenArmingCallback
        ] = None
        self._tamper_callback: Optional[SensorTamperCallback] = None
        self._door_open_when_arming = False
        self._proto_idx = proto_idx
        self._extra_data: Any = None
        self._unavailable = False

        self._definition: Optional[SensorDefinition] = None
        # Get sensor definition corresponds to the sensor type/subtype if any
        for s_def in SENSOR_DEFINITIONS:
            if (
                s_def.type == self._protocol_data.type_id
                and s_def.subtype == self._protocol_data.subtype  # noqa:W503
            ):
                self._definition = s_def
                break

    def update(self, obj: G90Sensor) -> None:
        """
        Updates sensor from another instance.

        :param obj: Sensor instance to update from
        """
        self._protocol_data = obj.protocol_data
        self._proto_idx = obj.proto_idx

    @property
    def name(self) -> str:
        """
        Sensor name, accounting for multi-channel entities (single
        protocol entity results in multiple :class:`.G90Sensor` instances).

        :return: Sensor name
        """
        if self._protocol_data.node_count == 1:
            return self._protocol_data.parent_name
        return f'{self._protocol_data.parent_name}#{self._subindex + 1}'

    @property
    def state_callback(self) -> Optional[SensorStateCallback]:
        """
        Callback that is invoked when the sensor changes its state.

        :return: Sensor state callback
        """
        return self._state_callback

    @state_callback.setter
    def state_callback(self, value: SensorStateCallback) -> None:
        self._state_callback = value

    @property
    def low_battery_callback(self) -> Optional[SensorLowBatteryCallback]:
        """
        Callback that is invoked when the sensor reports on low battery
        condition.

        :return: Sensor's low battery callback
        """
        return self._low_battery_callback

    @low_battery_callback.setter
    def low_battery_callback(self, value: SensorLowBatteryCallback) -> None:
        self._low_battery_callback = value

    @property
    def door_open_when_arming_callback(
        self
    ) -> Optional[SensorDoorOpenWhenArmingCallback]:
        """
        Callback that is invoked when the sensor reports on open door
        condition when arming.

        :return: Sensor's door open when arming callback
        """
        return self._door_open_when_arming_callback

    @door_open_when_arming_callback.setter
    def door_open_when_arming_callback(
        self, value: SensorDoorOpenWhenArmingCallback
    ) -> None:
        self._door_open_when_arming_callback = value

    @property
    def tamper_callback(self) -> Optional[SensorTamperCallback]:
        """
        Callback that is invoked when the sensor reports being tampered.

        :return: Sensor's tamper callback
        """
        return self._tamper_callback

    @tamper_callback.setter
    def tamper_callback(self, value: SensorTamperCallback) -> None:
        self._tamper_callback = value

    @property
    def occupancy(self) -> bool:
        """
        Occupancy (occupied/not occupied, or triggered/not triggered)
        for the sensor.

        :return: Sensor occupancy
        """
        return self._occupancy

    def _set_occupancy(self, value: bool) -> None:
        """
        Sets occupancy state of the sensor.
        Intentionally private, as occupancy state is derived from
        notifications/alerts.

        :param value: Occupancy state
        """
        _LOGGER.debug(
            "Setting occupancy for sensor index=%s: '%s' %s"
            " (previous value: %s)",
            self.index, self.name, value, self._occupancy
        )
        self._occupancy = value

    @property
    def protocol(self) -> G90SensorProtocols:
        """
        Protocol type of the sensor.

        :return: Protocol type
        """
        return G90SensorProtocols(self._protocol_data.protocol_id)

    @property
    def type(self) -> G90SensorTypes:
        """
        Type of the sensor.

        :return: Sensor type
        """
        return G90SensorTypes(self._protocol_data.type_id)

    @property
    def subtype(self) -> int:
        """
        Sub-type of the sensor.

        :return: Sensor sub-type
        """
        return self._protocol_data.subtype

    @property
    def reserved(self) -> G90SensorReservedFlags:
        """
        Reserved flags (read/write mode) for the sensor.

        :return: Reserved flags
        """
        return G90SensorReservedFlags(self._protocol_data.reserved_data)

    @property
    def user_flag(self) -> G90SensorUserFlags:
        """
        User flags for the sensor (disabled/enabled, arming type etc).

        :return: User flags
        """
        return G90SensorUserFlags(self._protocol_data.user_flag_data)

    @property
    def node_count(self) -> int:
        """
        Number of nodes (channels) for the sensor.

        :return: Number of nodes
        """
        return self._protocol_data.node_count

    @property
    def parent(self) -> G90Alarm:
        """
        Parent instance of alarm panel class the sensor is associated
        with.

        :return: Parent instance
        """
        return self._parent

    @property
    def index(self) -> int:
        """
        Index (internal position) of the sensor in the alarm panel.

        :return: Internal sensor position
        """
        return self._protocol_data.index

    @property
    def subindex(self) -> int:
        """
        Index of the sensor within multi-node device.

        :return: Index of sensor in multi-node device.
        """
        return self._subindex

    @property
    def proto_idx(self) -> int:
        """
        Index of the sensor within list of sensors as retrieved from the alarm
        panel.

        :return: Index of sensor in list of sensors.
        """
        return self._proto_idx

    @property
    def supports_enable_disable(self) -> bool:
        """
        Indicates if disabling/enabling the sensor is supported.

        :return: Support for enabling/disabling the sensor
        """
        return self._definition is not None

    @property
    def protocol_data(self) -> G90SensorIncomingData:
        """
        Protocol data of the sensor.

        :return: Protocol data
        """
        return self._protocol_data

    @property
    def is_wireless(self) -> bool:
        """
        Indicates if the sensor is wireless.
        """
        return self.protocol not in (G90SensorProtocols.CORD,)

    @property
    def is_low_battery(self) -> bool:
        """
        Indicates if the sensor is reporting low battery.

        The condition is cleared when the sensor reports activity (i.e. is no
        longer low on battery as it is able to report the activity).
        """
        return self._low_battery

    def _set_low_battery(self, value: bool) -> None:
        """
        Sets low battery state of the sensor.

        Intentionally private, as low battery state is derived from
        notifications/alerts.

        :param value: Low battery state
        """
        _LOGGER.debug(
            "Setting low battery for sensor index=%s '%s': %s"
            " (previous value: %s)",
            self.index, self.name, value, self._low_battery
        )
        self._low_battery = value

    @property
    def is_tampered(self) -> bool:
        """
        Indicates if the sensor has been tampered.

        The condition is cleared when panel is armed/disarmed next time.
        """
        return self._tampered

    def _set_tampered(self, value: bool) -> None:
        """
        Sets tamper state of the sensor.

        Intentionally private, as tamper state is derived from
        notifications/alerts.

        :param value: Tamper state
        """
        _LOGGER.debug(
            "Setting tamper for sensor index=%s '%s': %s"
            " (previous value: %s)",
            self.index, self.name, value, self._tampered
        )
        self._tampered = value

    @property
    def is_door_open_when_arming(self) -> bool:
        """
        Indicates if the sensor reports on open door when arming.

        The condition is cleared when panel is armed/disarmed next time.
        """
        return self._door_open_when_arming

    def _set_door_open_when_arming(self, value: bool) -> None:
        """
        Sets door open state of the sensor when arming.

        Intentionally private, as door open state is derived from
        notifications/alerts.

        :param value: Door open state
        """
        _LOGGER.debug(
            "Setting door open when arming for sensor index=%s '%s': %s"
            " (previous value: %s)",
            self.index, self.name, value, self._door_open_when_arming
        )
        self._door_open_when_arming = value

    @property
    def enabled(self) -> bool:
        """
        Indicates if the sensor is enabled.

        :return: If sensor is enabled
        """
        return self.user_flag & G90SensorUserFlags.ENABLED != 0

    async def set_user_flag(self, value: G90SensorUserFlags) -> None:
        """
        Sets user flags of the sensor.

        :param value: User flags to set, values other than
          :attr:`.G90SensorUserFlags.USER_SETTABLE` will be ignored and
          preserved from existing sensor flags.
        """
        if value & ~G90SensorUserFlags.USER_SETTABLE:
            _LOGGER.warning(
                'User flags for sensor index=%s contain non-user settable'
                ' flags, those will be ignored: %s',
                self.index, repr(value & ~G90SensorUserFlags.USER_SETTABLE)
            )
        # Checking private attribute directly, since `mypy` doesn't recognize
        # the check for sensor definition to be defined if done over
        # `self.supports_enable_disable` property
        if not self._definition:
            _LOGGER.warning(
                'Manipulating with user flags for sensor index=%s'
                ' is unsupported - no sensor definition for'
                ' type=%s, subtype=%s',
                self.index,
                self._protocol_data.type_id,
                self._protocol_data.subtype
            )

            return

        # Refresh actual sensor data from the alarm panel before modifying it.
        # This implies the sensor is at the same position within sensor list
        # (`_proto_index`) as it has been read initially from the alarm panel
        # when instantiated.
        _LOGGER.debug(
            'Refreshing sensor at index=%s, position in protocol list=%s',
            self.index, self.proto_idx
        )
        sensors_result = self.parent.paginated_result(
            G90Commands.GETSENSORLIST,
            start=self.proto_idx, end=self.proto_idx
        )
        sensors = [x.data async for x in sensors_result]

        # Abort if sensor is not found
        if not sensors:
            _LOGGER.error(
                'Sensor index=%s not found when attempting to set its'
                ' user flag',
                self.index,
            )
            return

        # Compare actual sensor data from what the sensor has been instantiated
        # from, and abort the operation if out-of-band changes are detected.
        sensor_data = sensors[0]
        if self._protocol_incoming_data_kls(
            *sensor_data
        ) != self._protocol_data:
            _LOGGER.error(
                "Sensor index=%s '%s' has been changed externally,"
                " refusing to alter its user flag",
                self.index,
                self.name
            )
            return

        prev_user_flag = self.user_flag

        # Re-instantiate the protocol data with modified user flags
        _data = asdict(self._protocol_data)
        _data['user_flag_data'] = (
            # Preserve flags that are not user-settable
            self.user_flag & ~G90SensorUserFlags.USER_SETTABLE
        ) | (
            # Combine them with the new user-settable flags
            value & G90SensorUserFlags.USER_SETTABLE
        )
        self._protocol_data = self._protocol_incoming_data_kls(**_data)

        _LOGGER.debug(
            'Sensor index=%s: previous user_flag %s, resulting user_flag %s',
            self._protocol_data.index,
            repr(prev_user_flag),
            repr(self.user_flag)
        )

        # Generate protocol data from write operation, deriving values either
        # from fields read from the sensor, or from the sensor definition - not
        # all fields are present during read, only in definition.
        outgoing_data = self._protocol_outgoing_data_kls(
            parent_name=self._protocol_data.parent_name,
            index=self._protocol_data.index,
            room_id=self._protocol_data.room_id,
            type_id=self._protocol_data.type_id,
            subtype=self._protocol_data.subtype,
            timeout=self._protocol_data.timeout,
            user_flag_data=self._protocol_data.user_flag_data,
            baudrate=self._protocol_data.baudrate,
            protocol_id=self._protocol_data.protocol_id,
            reserved_data=self._definition.reserved_data,
            node_count=self._protocol_data.node_count,
            rx=self._definition.rx,
            tx=self._definition.tx,
            private_data=self._definition.private_data,
        )
        # Modify the sensor
        await self._parent.command(
            G90Commands.SETSINGLESENSOR, list(astuple(outgoing_data))
        )

    async def set_enabled(self, value: bool) -> None:
        """
        Sets the sensor enabled/disabled.
        """

        # Modify the value of the user flag setting enabled/disabled one
        # appropriately.
        user_flag = self.user_flag
        if value:
            user_flag |= G90SensorUserFlags.ENABLED
        else:
            user_flag &= ~G90SensorUserFlags.ENABLED
        await self.set_user_flag(user_flag)

    @property
    def extra_data(self) -> Any:
        """
        Extra data for the sensor, that can be used to store
        caller-specific information and will be carried by the sensor instance.
        """
        return self._extra_data

    @extra_data.setter
    def extra_data(self, val: Any) -> None:
        self._extra_data = val

    @property
    def is_unavailable(self) -> bool:
        """
        Indicates if the sensor is unavailable (e.g. has been removed).
        """
        return self._unavailable

    @is_unavailable.setter
    def is_unavailable(self, value: bool) -> None:
        self._unavailable = value

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns dictionary representation of the sensor.

        :return: Dictionary representation
        """
        return {
            'name': self.name,
            'type': self.type,
            'subtype': self.subtype,
            'index': self.index,
            'protocol_index': self.proto_idx,
            'subindex': self.subindex,
            'node_count': self.node_count,
            'protocol': self.protocol,
            'occupancy': self.occupancy,
            'user_flag': self.user_flag,
            'reserved': self.reserved,
            'extra_data': self.extra_data,
            'enabled': self.enabled,
            'supports_enable_disable': self.supports_enable_disable,
            'is_wireless': self.is_wireless,
            'is_low_battery': self.is_low_battery,
            'is_tampered': self.is_tampered,
            'is_door_open_when_arming': self.is_door_open_when_arming,
            'is_unavailable': self.is_unavailable,
        }

    def __repr__(self) -> str:
        """
        Returns string representation of the sensor.

        :return: String representation
        """
        return super().__repr__() + f'({repr(self._asdict())})'

    def __eq__(self, value: object) -> bool:
        """
        Compares the sensor with another object.

        :param value: Object to compare with
        :return: If the sensor is equal to the object
        """
        if not isinstance(value, G90Sensor):
            return False

        return (
            self.type == value.type
            and self.subtype == value.subtype
            and self.index == value.index
            and self.name == value.name
        )



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements support for notifications/alerts sent by G90 alarm panel.
"""
import json
import logging
from typing import (
    Optional, List, Any, Callable
)
from dataclasses import dataclass
from asyncio.transports import BaseTransport
from datetime import datetime, timezone

from ..callback import G90Callback
from ..const import (
    G90MessageTypes,
    G90NotificationTypes,
    G90AlertTypes,
    G90AlertStateChangeTypes,
    G90ArmDisarmTypes,
    G90AlertSources,
    G90AlertStates,
    G90RemoteButtonStates,
)
from .protocol import G90NotificationProtocol

_LOGGER = logging.getLogger(__name__)


@dataclass
class G90Message:
    """
    Represents the message received from the device.

    :meta private:
    """
    code: G90MessageTypes
    data: List[Any]


@dataclass
class G90Notification:
    """
    Represents the notification received from the device.

    :meta private:
    """
    kind: G90NotificationTypes
    data: List[Any]


@dataclass
class G90ZoneInfo:
    """
    Represents zone details received from the device.

    :meta private:
    """
    idx: int
    name: str


@dataclass
class G90ArmDisarmInfo:
    """
    Represents the arm/disarm state received from the device.

    :meta private:
    """
    state: int


@dataclass
class G90DeviceAlert:  # pylint: disable=too-many-instance-attributes
    """
    Represents alert received from the device.
    """
    type: G90AlertTypes
    event_id: G90AlertStateChangeTypes
    source: G90AlertSources
    state: int
    zone_name: str
    device_id: str
    unix_time: int
    resv4: int
    other: str


class G90NotificationsBase:
    """
    Implements support for notifications/alerts sent by alarm panel.

    There is a basic check to ensure only notifications/alerts from the correct
    device are processed - the check uses the host and port of the device, and
    the device ID (GUID) that is set by the ancestor class that implements the
    commands (e.g. :class:`G90Alarm`). The latter to work correctly needs a
    command to be performed first, one that fetches device GUID and then stores
    it using :attr:`.device_id` (e.g. :meth:`G90Alarm.get_host_info`).
    """
    def __init__(
        self, protocol_factory: Callable[[], G90NotificationProtocol],
    ):
        # pylint: disable=too-many-arguments
        self._transport: Optional[BaseTransport] = None
        # Same but for device ID (GUID) - the notifications logic uses it to
        # perform validation, but doesn't set it from messages received (it
        # will diminish the purpose of the validation, should be done by an
        # ancestor class).
        self._device_id: Optional[str] = None
        self._protocol = protocol_factory()
        self._last_device_packet_time: Optional[datetime] = None
        self._last_upstream_packet_time: Optional[datetime] = None

    def handle_notification(
        self, notification: G90Notification
    ) -> None:
        """
        Handles notification received from the device.

        :param notification: The notification to handle.
        """
        # Sensor activity notification
        if notification.kind == G90NotificationTypes.SENSOR_ACTIVITY:
            g90_zone_info = G90ZoneInfo(*notification.data)

            _LOGGER.debug('Sensor notification: %s', g90_zone_info)
            G90Callback.invoke(
                self._protocol.on_sensor_activity,
                g90_zone_info.idx, g90_zone_info.name
            )

            return

        # Arm/disarm notification
        if notification.kind == G90NotificationTypes.ARM_DISARM:
            g90_armdisarm_info = G90ArmDisarmInfo(
                *notification.data)
            # Map the state received from the device to corresponding enum
            state = G90ArmDisarmTypes(g90_armdisarm_info.state)

            _LOGGER.debug('Arm/disarm notification: %s',
                          state)
            G90Callback.invoke(self._protocol.on_armdisarm, state)

            return

        # An open door is detected when arming
        if notification.kind == G90NotificationTypes.DOOR_OPEN_WHEN_ARMING:
            g90_zone_info = G90ZoneInfo(*notification.data)
            _LOGGER.debug('Door open detected when arming: %s', g90_zone_info)
            G90Callback.invoke(
                self._protocol.on_door_open_when_arming,
                g90_zone_info.idx, g90_zone_info.name
            )
            return

        _LOGGER.warning('Unknown notification received:'
                        ' kind %s, data %s',
                        notification.kind, notification.data)

    def _handle_alert_sensor_activity(self, alert: G90DeviceAlert) -> bool:
        """
        Handles sensor activity alert.
        """
        if alert.source == G90AlertSources.REMOTE:
            _LOGGER.debug('Remote button press alert: %s', alert)
            G90Callback.invoke(
                self._protocol.on_remote_button_press,
                alert.event_id, alert.zone_name,
                G90RemoteButtonStates(alert.state)
            )

            return True

        if alert.state in (
            G90AlertStates.DOOR_OPEN, G90AlertStates.DOOR_CLOSE
        ):
            is_open = (
                alert.source == G90AlertSources.SENSOR
                and alert.state == G90AlertStates.DOOR_OPEN  # noqa: W503
            ) or alert.source == G90AlertSources.DOORBELL

            _LOGGER.debug('Door open_close alert: %s', alert)
            G90Callback.invoke(
                self._protocol.on_door_open_close,
                alert.event_id, alert.zone_name, is_open
            )

            return True

        if (
            alert.source == G90AlertSources.SENSOR
            and alert.state == G90AlertStates.LOW_BATTERY  # noqa: W503
        ):
            _LOGGER.debug('Low battery alert: %s', alert)
            G90Callback.invoke(
                self._protocol.on_low_battery,
                alert.event_id, alert.zone_name
            )

            return True

        return False

    def handle_alert(
        self, alert: G90DeviceAlert,
        verify_device_id: bool = True
    ) -> None:
        """
        Handles alert received from the device.

        :param alert: The alert to handle.
        :param verify_device_id: Whether to verify the device ID (GUID) in the
         alert. If set to False, the device ID will not be verified.
        """
        handled = False

        # Stop processing when alert is received from the device with different
        # GUID (if enabled)
        if (
            verify_device_id and self.device_id
            and alert.device_id != self.device_id
        ):
            _LOGGER.error(
                "Received alert from wrong device: expected '%s', got '%s'",
                self.device_id, alert.device_id
            )
            return

        if alert.type == G90AlertTypes.SENSOR_ACTIVITY:
            handled = self._handle_alert_sensor_activity(alert)

        if alert.type == G90AlertTypes.STATE_CHANGE:
            # Define the mapping between device state received in the alert, to
            # common `G90ArmDisarmTypes` enum that is used when setting device
            # arm state and received in the corresponding notifications. The
            # primary reason is to unify state as passed down to the callbacks.
            # The map covers only subset of state changes pertinent to
            # arm/disarm state changes
            alarm_arm_disarm_state_map = {
                G90AlertStateChangeTypes.ARM_HOME: G90ArmDisarmTypes.ARM_HOME,
                G90AlertStateChangeTypes.ARM_AWAY: G90ArmDisarmTypes.ARM_AWAY,
                G90AlertStateChangeTypes.DISARM: G90ArmDisarmTypes.DISARM
            }

            state = alarm_arm_disarm_state_map.get(alert.event_id, None)
            if state:
                # We received the device state change related to arm/disarm,
                # invoke the corresponding callback
                _LOGGER.debug('Arm/disarm state change: %s', state)
                G90Callback.invoke(self._protocol.on_armdisarm, state)

            handled = True

        if alert.type == G90AlertTypes.ALARM:
            # Remote SOS
            if alert.source == G90AlertSources.REMOTE:
                _LOGGER.debug('SOS: %s', alert.zone_name)
                G90Callback.invoke(
                    self._protocol.on_sos,
                    alert.event_id, alert.zone_name, False
                )
            # Regular alarm
            else:
                is_tampered = alert.state == G90AlertStates.TAMPER
                _LOGGER.debug(
                    'Alarm: %s, is tampered: %s', alert.zone_name, is_tampered
                )
                G90Callback.invoke(
                    self._protocol.on_alarm,
                    alert.event_id, alert.zone_name, is_tampered
                )

            handled = True

        # Host SOS
        if alert.type == G90AlertTypes.HOST_SOS:
            zone_name = 'Host SOS'

            _LOGGER.debug('SOS: Host')
            G90Callback.invoke(
                self._protocol.on_sos, alert.event_id, zone_name, True
            )

            handled = True

        if not handled:
            _LOGGER.warning(
                'Unknown alert received: type %s, data %s',
                alert.type, alert
            )

    # pylint:disable=too-many-return-statements
    def handle(self, data: bytes) -> None:
        """
        Invoked when message is received from the device.
        """
        try:
            s_data = data.decode('utf-8')
        except UnicodeDecodeError:
            _LOGGER.error('Unable to decode device message from UTF-8')
            return

        if not s_data.endswith('\0'):
            _LOGGER.error('Missing end marker in data')
            return
        payload = s_data[:-1]

        try:
            message = json.loads(payload)
            g90_message = G90Message(*message)
        except json.JSONDecodeError as exc:
            _LOGGER.error("Unable to parse device message '%s' as JSON: %s",
                          payload, exc)
            return
        except TypeError as exc:
            _LOGGER.error("Device message '%s' is malformed: %s",
                          payload, exc)
            return

        # Device notifications
        if g90_message.code == G90MessageTypes.NOTIFICATION:
            try:
                notification_data = G90Notification(*g90_message.data)
            except TypeError as exc:
                _LOGGER.error('Bad notification received: %s', exc)
                return
            self.handle_notification(notification_data)
            return

        # Device alerts
        if g90_message.code == G90MessageTypes.ALERT:
            try:
                alert_data = G90DeviceAlert(*g90_message.data)
            except TypeError as exc:
                _LOGGER.error('Bad alert received: %s', exc)
                return
            self.handle_alert(alert_data)
            return

        _LOGGER.warning('Unknown message received: %s', message)

    async def listen(self) -> None:
        """
        Listens for notifications/alerts from the device.
        """
        raise NotImplementedError

    @property
    def listener_started(self) -> bool:
        """
        Indicates if the listener of the device notifications has been started.
        """
        return self._transport is not None

    async def close(self) -> None:
        """
        Closes the listener.
        """
        if self._transport:
            _LOGGER.debug('No longer listening for device notifications')
            self._transport.close()
            self._transport = None

    @property
    def device_id(self) -> Optional[str]:
        """
        The ID (GUID) of the panel being communicated with thru commands.

        Available when any panel command receives it from the device
        (`GETHOSTINFO` local command or Hello / HelloDiscovery cloud ones).
        """
        return self._device_id

    @device_id.setter
    def device_id(self, device_id: str) -> None:
        # Under not yet identified circumstances the device ID might be empty
        # string provided by :meth:`G90Alarm.get_host_info` - disallow that
        if not device_id or len(device_id.strip()) == 0:
            _LOGGER.debug(
                'Device ID is empty or contains whitespace only, not setting'
            )
            return

        self._device_id = device_id

    def clear_device_id(self) -> None:
        """
        Clears the device ID.
        """
        self._device_id = None

    @property
    def last_device_packet_time(self) -> Optional[datetime]:
        """
        Returns the timestamp of the last packet received from the device.

        This property can be used to monitor the communication health with the
        device.
        """
        return self._last_device_packet_time

    def set_last_device_packet_time(self) -> None:
        """
        Updates the timestamp of the last packet received from the device.

        This method is called internally when a packet is received from the
        device.
        """
        self._last_device_packet_time = datetime.now(tz=timezone.utc)
        _LOGGER.debug(
            'Last device packet time: %s', self._last_device_packet_time
        )

    @property
    def last_upstream_packet_time(self) -> Optional[datetime]:
        """
        Returns the timestamp of the last packet sent to the upstream server.

        This property can be used to monitor the communication health with the
        cloud/upstream server.
        """
        return self._last_upstream_packet_time

    def set_last_upstream_packet_time(self) -> None:
        """
        Updates the timestamp of the last packet sent to the upstream server.

        This method is called internally when a packet is sent to the upstream
        server.
        """
        self._last_upstream_packet_time = datetime.now(tz=timezone.utc)
        _LOGGER.debug(
            'Last upstream packet time: %s', self._last_upstream_packet_time
        )



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Constants used in the G90 cloud protocol implementation.

This module defines the main enumerations for direction and command types
used in the cloud protocol communication with G90 alarm systems.
"""
from enum import IntEnum


class G90CloudDirection(IntEnum):
    """
    Defines the direction of G90 cloud protocol messages.

    These values indicate whether messages are flowing from the cloud to the
    device, from the device to the cloud, or are part of discovery processes.
    """
    UNSPECIFIED = 0
    CLOUD = 32  # 0x20
    DEVICE = 16  # 0x10
    DEVICE_DISCOVERY = 48  # 0x30
    CLOUD_DISCOVERY = 208  # 0xD0


class G90CloudCommand(IntEnum):
    """
    Defines the command types used in G90 cloud protocol messages.

    These values identify the purpose of each cloud protocol message,
    such as hello messages, notifications, commands, and status updates.
    """
    HELLO = 0x01
    HELLO_ACK = 0x41
    NOTIFICATION = 0x22
    STATUS_CHANGE = 0x21
    HELLO_INFO = 0x63
    COMMAND = 0x29



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Discovers G90 alarm panel devices with specific ID.
"""
from __future__ import annotations
import logging
from typing import Tuple, Any, Optional, Dict, List
from dataclasses import dataclass, asdict
import asyncio
from asyncio.transports import BaseTransport
from .base_cmd import G90BaseCommand
from ..const import G90Commands
from ..exceptions import G90Error

_LOGGER = logging.getLogger(__name__)


@dataclass
# pylint: disable=too-many-instance-attributes
class G90TargetedDiscoveryInfo:
    """
    Wire representation of the information about discovered device.
    """
    message: str
    product_name: str
    wifi_protocol_version: str
    cloud_protocol_version: str
    mcu_hw_version: str
    fw_version: str
    gsm_status: str
    wifi_status: str
    server_status: str
    reserved1: str
    reserved2: str
    gsm_signal_level: str
    wifi_signal_level: str

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the information about discovered device as dictionary.
        """
        return asdict(self)


@dataclass
class G90DiscoveredDeviceTargeted(G90TargetedDiscoveryInfo):
    """
    Discovered device with specific ID.
    """
    host: str
    port: int
    guid: str


class G90TargetedDiscovery(G90BaseCommand):
    """
    Discovers alarm panel devices with specific ID.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, device_id: str, **kwargs: Any):
        super().__init__(
            # No actual command will be processed by base class, `NONE` is used
            # for proper typing only
            code=G90Commands.NONE, **kwargs
        )
        self._device_id = device_id
        self._discovered_devices: List[G90DiscoveredDeviceTargeted] = []

    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection is established.
        """

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Invoked when connection is lost.
        """

    # Implementation of datagram protocol,
    # https://docs.python.org/3/library/asyncio-protocol.html#datagram-protocols
    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """
        Invoked when datagram is received.
        """
        try:
            _LOGGER.debug('Received from %s:%s: %s', addr[0], addr[1], data)
            try:
                decoded = data.decode('utf-8')
            except UnicodeDecodeError as exc:
                raise G90Error(
                    'Unable to decode discovery response from UTF-8'
                ) from exc
            if not decoded.endswith('\0'):
                raise G90Error('Invalid discovery response')
            host_info = G90TargetedDiscoveryInfo(*decoded[:-1].split(','))
            if host_info.message != 'IWTAC_PROBE_DEVICE_ACK':
                raise G90Error('Invalid discovery response')
            res = G90DiscoveredDeviceTargeted(
                host=addr[0],
                port=addr[1],
                guid=self._device_id,
                **host_info._asdict()
            )
            _LOGGER.debug('Discovered device: %s', res)
            self.add_device(res)
        except Exception as exc:  # pylint: disable=broad-except
            _LOGGER.warning('Got exception, ignoring: %s', exc)

    def error_received(self, exc: Exception) -> None:
        """
        Invoked when error is received.
        """

    def to_wire(self) -> bytes:
        """
        Converts the command to wire representation.
        """
        return bytes(f'IWTAC_PROBE_DEVICE,{self._device_id}\0', 'ascii')

    async def process(self) -> G90TargetedDiscovery:
        """
        Initiates the device discovery process.
        """
        _LOGGER.debug('Attempting device discovery...')
        transport, _ = await self._create_connection()
        transport.sendto(self.to_wire())
        await asyncio.sleep(self._timeout)
        transport.close()
        _LOGGER.debug('Discovered %s devices', len(self.devices))
        return self

    @property
    def devices(self) -> List[G90DiscoveredDeviceTargeted]:
        """
        The list of discovered devices.
        """
        return self._discovered_devices

    def add_device(self, value: G90DiscoveredDeviceTargeted) -> None:
        """
        Adds discovered device to the list.
        """
        self._discovered_devices.append(value)



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Base entity list.
"""
from abc import ABC, abstractmethod
from typing import (
    List, AsyncGenerator, Optional, TypeVar, Generic, cast, TYPE_CHECKING,
)
import asyncio
import logging

from .base_entity import G90BaseEntity
if TYPE_CHECKING:
    from ..alarm import G90Alarm
else:
    # Alias G90Alarm to object avoid circular imports
    # (`G90Alarm` -> `G90SensorList` -> `G90BaseList` -> `G90Alarm`)
    G90Alarm = object

T = TypeVar('T', bound=G90BaseEntity)
_LOGGER = logging.getLogger(__name__)


class G90BaseList(Generic[T], ABC):
    """
    Base entity list class.

    :param parent: Parent alarm panel instance.
    """
    def __init__(self, parent: G90Alarm) -> None:
        self._entities: List[T] = []
        self._lock = asyncio.Lock()
        self._parent = parent

    @abstractmethod
    async def _fetch(self) -> AsyncGenerator[T, None]:
        """
        Fetch the list of entities from the panel.

        :return: Async generator of entities
        """
        yield cast(T, None)

    @property
    async def entities(self) -> List[T]:
        """
        Return the list of entities.

        :meth:`update` is called if the list is empty.

        :return: List of entities
        """
        # Please see below for the explanation of the lock usage
        async with self._lock:
            entities = self._entities

        if not entities:
            return await self.update()

        return entities

    async def update(self) -> List[T]:
        """
        Update the list of entities from the panel.

        :return: List of entities
        """
        # Use lock around the operation, to ensure no duplicated entries in the
        # resulting list or redundant exchanges with panel are made when the
        # method is called concurrently
        async with self._lock:
            entities = self._fetch()

            non_existing_entities = self._entities.copy()
            async for entity in entities:
                try:
                    existing_entity_idx = self._entities.index(entity)
                except ValueError:
                    existing_entity_idx = None

                if existing_entity_idx is not None:
                    existing_entity = self._entities[existing_entity_idx]
                    # Update the existing entity with the new data
                    _LOGGER.debug(
                        "Updating existing entity '%s' from protocol"
                        " data '%s'", existing_entity, entity
                    )

                    self._entities[existing_entity_idx].update(entity)
                    non_existing_entities.remove(entity)
                else:
                    # Add the new entity to the list
                    _LOGGER.debug('Adding new entity: %s', entity)
                    self._entities.append(entity)

            # Mark the entities that are no longer in the list
            for unavailable_entity in non_existing_entities:
                _LOGGER.debug(
                    'Marking entity as unavailable: %s', unavailable_entity
                )
                unavailable_entity.is_unavailable = True

            _LOGGER.debug(
                'Total number of entities: %s, unavailable: %s',
                len(self._entities), len(non_existing_entities)
            )

            return self._entities

    async def find(
        self, idx: int, name: str, exclude_unavailable: bool
    ) -> Optional[T]:
        """
        Finds entity by index and name.

        :param idx: Entity index
        :param name: Entity name
        :param exclude_unavailable: Exclude unavailable entities
        :return: Entity instance
        """
        entities = await self.entities

        found = None
        # Fast lookup by direct index
        if idx < len(entities) and entities[idx].name == name:
            entity = entities[idx]
            _LOGGER.debug('Found entity via fast lookup: %s', entity)
            found = entity

        # Fast lookup failed, perform slow one over the whole entities list
        if not found:
            for entity in entities:
                if entity.index == idx and entity.name == name:
                    _LOGGER.debug('Found entity: %s', entity)
                    found = entity

        if found:
            if not exclude_unavailable or not found.is_unavailable:
                return found

            _LOGGER.debug(
                'Entity is found but unavailable, will result in none returned'
            )

        _LOGGER.error('Entity not found: idx=%s, name=%s', idx, name)
        return None
